import json
from http.cookies import SimpleCookie

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QDialog, QTextEdit, QPushButton, \
    QVBoxLayout, QLabel
try:
    from feeluown.gui.widgets.weblogin import WebLoginView
except ImportError:
    has_webengine = False
else:
    has_webengine = True

from feeluown.utils import aio


class InvalidCookies(Exception):
    pass


class LoginDialog(QDialog):
    """Base class for login dialogs"""
    #: login succeed signal
    login_succeed = pyqtSignal()


class CookiesLoginDialog(LoginDialog):
    """

    CookiesLoginDialog provides a text edit area and a login button.
    User firstly fill in the text edit area with cookies. User can
    then click the login button. The `clicked` signal is connected to
    :meth:`.login`.

    Cookies can be in text or json format. User can copy cookies from
    web browser like Chrome or Firefox. Cookies copied from firefox are
    in json format and cookies copied from Chrome are in text format.

    Subclass MUST implement four methods.

    - :meth:`setup_user`
    - :meth:`user_from_cookies`
    - :meth:`load_user_cookies`
    - :meth:`dump_user_cookies`

    One usage example: feeluown-qqmusic.
    """

    def __init__(self, uri: str = None, required_cookies_fields=None):
        if has_webengine and uri and required_cookies_fields:
            use_webview = True
            flags = Qt.Window
        else:
            use_webview = False
            flags = Qt.Popup

        super().__init__(None, flags)
        self._use_webview = use_webview
        self._uri = uri
        self._required_cookies_fields = required_cookies_fields

        self.cookies_text_edit = QTextEdit(self)
        self.hint_label = QLabel(self)
        self.login_btn = QPushButton('登录', self)
        self.weblogin_btn = QPushButton('网页登录', self)

        self.hint_label.setTextFormat(Qt.RichText)

        self._layout = QVBoxLayout(self)
        self._layout.addWidget(self.cookies_text_edit)
        self._layout.addWidget(self.hint_label)
        self._layout.addWidget(self.login_btn)
        self._layout.addWidget(self.weblogin_btn)

        self.cookies_text_edit.setAcceptRichText(False)
        self.cookies_text_edit.setPlaceholderText(
            '请从浏览器中复制 Cookie：\n\n'
            'Chrome 复制的 cookie 格式类似：key1=value1; key2=value2\n\n'
            'Firefox 复制的 cookie 格式类似：{"key1": value1, "key1": value2}'
        )

        if self._use_webview is True:
            self.weblogin_btn.clicked.connect(self._start_web_login)
        else:
            # disable the button if feeluown does not support
            if uri and required_cookies_fields and not has_webengine:
                self.weblogin_btn.setEnabled(False)
            else:
                # hide the button if provider does not support
                self.weblogin_btn.hide()
        self.login_btn.clicked.connect(lambda: aio.create_task(self.login()))
        self.login_succeed.connect(self.hide)

    def _start_web_login(self):
        self._web_login = WebLoginView(self._uri, self._required_cookies_fields)
        self._web_login.succeed.connect(self._on_web_login_succeed)
        self._web_login.show()

    def _on_web_login_succeed(self, cookies):
        self.cookies_text_edit.setText(json.dumps(cookies, indent=2))
        self._web_login.close()
        del self._web_login
        aio.create_task(self.login())

    def _parse_json_cookies(self, text):
        try:
            return json.loads(text)
        except ValueError:
            return None

    def _parse_text_cookies(self, text):
        """
        cookies: uid=111; userAction=222;
        """
        cookie = SimpleCookie()
        cookie.load(text)
        cookie_dict = {}
        for _, morsel in cookie.items():
            cookie_dict[morsel.key] = morsel.value
        return cookie_dict or None

    def get_cookies(self):
        """Parse the content in text edit area

        :return: return None when the content is invalid
        :rtype: dict or None
        """
        # We assume users only use firefox and chrome. Cookies from firefox are
        # in json format. Cookies copied from chrome are in text format.
        parsers = (('json', self._parse_json_cookies),
                   ('text', self._parse_text_cookies))
        text = self.cookies_text_edit.toPlainText()
        cookies = None
        for name, parse in parsers:
            cookies = parse(text)
            if cookies is None:
                self.show_hint(f'使用 {name} 解析器解析失败，尝试下一种', color='orange')
            else:
                self.show_hint(f'使用 {name} 解析器解析成功')
                break
        return cookies

    def show_hint(self, text, color=None):
        """Show hint message on dialog

        :param string color: red for error, orange for warning,
            green for success
        """
        if color is None:
            color = ''
        self.hint_label.setText(f"<p style='color: {color}'>{text}</p>")

    async def login(self):
        """Login with cookies

        Read cookies that has been filled in and create a user from it.
        If succeed, the :attr:`.login_succeed` signal will be emit.
        If failed, show specific error message on the dialog based on
        the exception.
        """
        await self.login_with_cookies(self.get_cookies())

    async def login_with_cookies(self, cookies):
        try:
            user = await self.user_from_cookies(cookies)
        except InvalidCookies as e:
            self.show_hint(str(e), color='red')
        else:
            self.show_hint('保存用户信息到 FeelUOwn 数据目录')
            self.dump_user_cookies(user, cookies)
            self.setup_user(user)
            self.login_succeed.emit()

    def autologin(self):
        """Try to load user cookies and login with it

        Generally, you can call this method after dialog is shown.
        """
        cookies = self.load_user_cookies()
        if cookies is not None:
            self.show_hint('正在尝试加载已有用户...', color='green')
            self.cookies_text_edit.setText(json.dumps(cookies, indent=2))
            aio.create_task(self.login_with_cookies(cookies))

    def setup_user(self, user):
        """Setup user session

        :type user: feeluown.models.UserModel
        """
        raise NotImplementedError

    async def user_from_cookies(self, cookies):
        """Create a user model from cookies dict

        :type cookies: dict
        :rtype: feeluown.models.UserModel
        """
        raise NotImplementedError

    def load_user_cookies(self):
        """Load user cookies from somewhere

        Load the cookies that is dumped before. If the load processing failed,
        just return None.

        :return: cookies in dict format
        :rtype: dict or None
        """
        raise NotImplementedError

    def dump_user_cookies(self, user, cookies):
        """Dump user cookies to somewhere

        Generally, you can store the cookies in FeelUOwn data directory
        with specifical filename.

        :type user: feeluown.models.UserModel
        :type cookies: dict
        """
        raise NotImplementedError


if __name__ == '__main__':
    from PyQt5.QtWidgets import QApplication

    app = QApplication([])
    widget = CookiesLoginDialog()
    widget.show()
    widget.show_hint('格式可能不正确', color='orange')
    app.exec()



import logging
from contextlib import contextmanager
from enum import IntEnum

from feeluown.player.mpvplayer import _mpv_set_property_string
from feeluown.gui.widgets.frameless import ResizableFramelessContainer

logger = logging.getLogger(__name__)


class Mode(IntEnum):
    none = 0
    normal = 1
    pip = 2  # Picture in picture


class VideoShowCtl:
    def __init__(self, app, ui):
        """video show controller

        :type app: feeluown.app.App
        :type ui: feeluown.ui.Ui
        """
        self._app = app
        self._ui = ui
        self._pip_container = ResizableFramelessContainer()
        self._mode = Mode.none
        self._parent_is_normal = True
        self._parent_is_changing = False

        self._ui.pc_panel.mv_btn.clicked.connect(self.play_mv)
        self._ui.toggle_video_btn.clicked.connect(lambda: self.set_mode(Mode.normal))
        self._ui.pc_panel.toggle_pip_btn.clicked.connect(lambda: self.set_mode(Mode.pip))

        self._ui.mpv_widget.show()
        # Set aioqueue=True so that mpv_widget hide after it is shown once.
        self._app.initialized.connect(
            lambda _: self._ui.mpv_widget.hide(), weak=False, aioqueue=True)
        self._app.player.media_changed.connect(self.on_media_changed, aioqueue=True)
        self._app.player.video_format_changed.connect(
            self.on_video_format_changed, aioqueue=True)

        self._pip_container.setMinimumSize(200, 130)
        self._pip_container.hide()

    def show_ctl_btns(self):
        self._ui.toggle_video_btn.show()
        self._ui.pc_panel.toggle_pip_btn.show()

    def hide_ctl_btns(self):
        self._ui.toggle_video_btn.hide()
        self._ui.pc_panel.toggle_pip_btn.hide()

    @contextmanager
    def change_parent(self):
        self._parent_is_changing = True
        self._before_change_mpv_widget_parent()
        yield
        self._after_change_mpv_widget_parent()
        self._parent_is_changing = False

    def enter_normal_mode(self):
        """enter normal mode"""
        self._ui.bottom_panel.hide()
        self._ui._splitter.hide()
        self._ui.pc_panel.toggle_video_btn.show()
        self._ui.pc_panel.toggle_video_btn.setText('▽')
        logger.info("enter video-show normal mode")
        self._ui.mpv_widget.overlay_auto_visible = False
        self._ui.mpv_widget.keep_wh_ratio = False
        if self._parent_is_normal is False:
            with self.change_parent():
                self._ui.mpv_widget.hide()
                self._ui.mpv_widget.shutdown()
                self._pip_container.detach()
                self._app.layout().insertWidget(1, self._ui.mpv_widget)
                self._parent_is_normal = True
                self._ui.mpv_widget.show()
        else:
            self._ui.mpv_widget.show()

    def exit_normal_mode(self):
        self._ui.mpv_widget.hide()
        self._ui._splitter.show()
        self._ui.bottom_panel.show()
        self._ui.pc_panel.toggle_video_btn.setText('△')
        logger.info("exit video-show normal mode")

    def enter_pip_mode(self):
        """enter picture in picture mode"""
        self._ui.toggle_video_btn.hide()
        logger.info("enter video-show picture in picture mode")
        self._ui.mpv_widget.overlay_auto_visible = True
        if self._parent_is_normal is True:
            width = self._app.player._mpv.width
            height = self._app.player._mpv.height
            with self.change_parent():
                self._ui.mpv_widget.hide()
                self._ui.mpv_widget.shutdown()
                self._ui._splitter.show()
                self._ui.bottom_panel.show()
                self._app.layout().removeWidget(self._ui.mpv_widget)
                self._pip_container.attach_widget(self._ui.mpv_widget)
                self._parent_is_normal = False
                self._pip_container.show()
                self._ui.mpv_widget.show()
            proper_width = max(min(width, 640), 320)
            proper_height = height * proper_width // width
            self._pip_container.resize(proper_width, proper_height)
        else:
            self._pip_container.show()
            self._ui.mpv_widget.show()

    def exit_pip_mode(self):
        """exit picture in picture mode"""
        self._pip_container.hide()
        self._ui.mpv_widget.hide()
        logger.info("exit video-show picture in picture mode")

    #
    # signal callbacks
    #

    def play_mv(self):
        song = self._app.player.current_song
        if song is None:
            return

        # The mv button only shows when there is a valid mv object
        mv = self._app.library.song_get_mv(song)
        self._app.playlist.set_current_model(mv)

    def set_mode(self, mode):
        # change mode to none, exit orignal mode
        if mode is Mode.none:
            if self._mode is Mode.normal:
                self.exit_normal_mode()
            elif self._mode is Mode.pip:
                self.exit_pip_mode()
            self._mode = mode
            self.hide_ctl_btns()
            return
        # change current mode to mode
        #
        # if mode is same as the current mode, exit mode
        # if mode is not same as the current mode, exit
        # current mode and enter mode
        if self._mode is Mode.none:
            if mode is Mode.normal:
                self.enter_normal_mode()
            else:
                self.enter_pip_mode()
            self._mode = mode
        elif self._mode is Mode.normal:
            self.exit_normal_mode()
            self._mode = Mode.none
            if mode is Mode.pip:
                self.enter_pip_mode()
                self._mode = mode
        else:
            self.exit_pip_mode()
            self._mode = Mode.none
            if mode is Mode.normal:
                self.enter_normal_mode()
                self._mode = mode

    def on_media_changed(self, media):
        if not media:
            logger.info('Media is changed to empty, set mode to none')
            self.set_mode(Mode.none)

    def on_video_format_changed(self, video_format):
        # When video is available, show control buttons. If video
        # is unavailable, hide video and control buttons.
        if video_format is None:
            # ignore the signal if parent is changing
            if self._parent_is_changing:
                return
            logger.info('Media is changed to audio, set mode to none')
            self.set_mode(Mode.none)
        else:
            self.show_ctl_btns()

    #
    # private methods
    #
    def _before_change_mpv_widget_parent(self):
        """
        According to Qt docs, reparenting an OpenGLWidget will destory the GL context.
        In mpv widget, it calls _mpv_opengl_cb_uninit_gl. After uninit_gl, mpv can't show
        video anymore because video_out is destroyed.

        See mpv mpv_opengl_cb_uninit_gl implementation for more details.
        """
        _mpv_set_property_string(self._app.player._mpv.handle, b'vid', b'no')

    def _after_change_mpv_widget_parent(self):
        """
        To recover the video show, we should reinit gl and reinit video. gl is
        automatically reinited when the mpv_widget do painting. We should
        manually reinit video.

        NOTE(cosven): After some investigation, I found that the API in mpv to
        reinit video_out(maybe video is almost same as video_out)
        is init_best_video_out. Theoretically, sending 'video-reload' command
        will trigger this API. However, we can't run this command
        in our case and I don't know why. Another way to trigger
        the API is to switch track. Changing vid property just switch the track.

        Inpect mpv init_best_video_out caller for more details. You should see
        mp_switch_track_n is one of the entrypoint.
        """
        _mpv_set_property_string(self._app.player._mpv.handle, b'vid', b'1')



import logging
import sys

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QGuiApplication
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QLabel, QVBoxLayout, \
    QSizePolicy, QScrollArea, QFrame

from feeluown.models.uri import reverse, resolve
from feeluown.library import ProviderFlags as PF, NotSupported
from feeluown.player import parse_lyric_text
from feeluown.utils import aio
from feeluown.utils.reader import create_reader
from feeluown.gui.helpers import BgTransparentMixin, resize_font
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets.cover_label import CoverLabelV2
from feeluown.gui.widgets.comment_list import CommentListView, CommentListModel
from feeluown.gui.widgets.songs import SongListView, SongListModel

logger = logging.getLogger(__name__)


async def render(req, **kwargs):  # pylint: disable=too-many-locals
    app = req.ctx['app']
    song = req.ctx['model']

    # TODO: Initialize the view with song object, and it should reduce
    # the code complexity.
    view = SongExploreView(app=app)
    app.ui.right_panel.set_body(view)

    # show song album cover
    song = await aio.run_fn(app.library.song_upgrade, song)

    # bind signals
    view.play_btn.clicked.connect(lambda: app.playlist.play_model(song))
    if app.library.check_flags_by_model(song, PF.web_url):
        async def copy_song_web_url():
            web_url = await aio.run_fn(app.library.song_get_web_url, song)
            app.show_msg(f'已经复制：{web_url}')
            QGuiApplication.clipboard().setText(web_url)

        view.copy_web_url_btn.clicked.connect(lambda: aio.run_afn(copy_song_web_url))
        # TODO: Open url in browser when alt key is pressed. Use
        # QDesktopServices.openUrl to open url in browser, and
        # you may use QGuiApplication::keyboardModifiers to check
        # if alt key is pressed.
        #
        # NOTE(cosven): Since switching from applications is inconvenience,
        # the default behaviour of button is url-copy instead of url-open.
    else:
        view.copy_web_url_btn.hide()

    view.header_label.setText(f'<h1>{song.title}</h1>')
    aio.create_task(view.album_info_label.show_song(song))

    if app.library.check_flags_by_model(song, PF.similar):
        songs = await aio.run_fn(app.library.song_list_similar, song)
        model = SongListModel(create_reader(songs))
        view.similar_songs_view.setModel(model)
    else:
        view.similar_songs_header.setText('<span>该提供方暂不支持查看相似歌曲，555</span>')

    if app.library.check_flags_by_model(song, PF.hot_comments):
        comments = await aio.run_fn(app.library.song_list_hot_comments, song)
        comments_reader = create_reader(comments)
        view.comments_view.setModel(CommentListModel(comments_reader))
    else:
        view.comments_header.setText('<span>该提供方暂不支持查看歌曲评论，555</span>')

    try:
        lyric = app.library.song_get_lyric(song)
    except NotSupported as e:
        logger.info('cant show lyric due to %s', str(e))
    else:
        if lyric is not None:
            ms_sentence_map = parse_lyric_text(lyric.content)
            sentences = []
            for _, sentence in sorted(ms_sentence_map.items(),
                                      key=lambda item: item[0]):
                sentences.append(sentence)
            view.lyric_label.set_lyric('\n'.join(sentences))

    # Show album cover in the end since it's an expensive CPU/IO operation.
    # FIXME: handle NotSupported exception
    if song.album is not None:
        album = await aio.run_fn(app.library.album_upgrade, song.album)
        aio.create_task(view.cover_label.show_cover(album.cover, reverse(album)))


class ScrollArea(QScrollArea, BgTransparentMixin):
    def __init__(self, app, parent=None):
        super().__init__(parent=parent)

        self._app = app

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)

        # As far as I know, KDE and GNOME can't auto hide the scrollbar,
        # and they show an old-fation vertical scrollbar.
        # HELP: implement an auto-hide scrollbar for Linux
        if sys.platform.lower() != 'darwin':
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)


class HeaderLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setTextFormat(Qt.RichText)
        self.setTextInteractionFlags(Qt.TextSelectableByMouse)


class LyricLabel(QLabel, BgTransparentMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWordWrap(True)
        self.setTextInteractionFlags(Qt.TextSelectableByMouse)
        font = self.font()
        resize_font(font, -1)
        self.setFont(font)
        self.setTextFormat(Qt.RichText)

    def set_lyric(self, text):
        text = text.replace("\n", "<br>")
        self.setText(f'<p style="line-height: 120%">{text}</p>')


class SongAlbumLabel(QLabel):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app
        self.setTextFormat(Qt.RichText)
        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.linkActivated.connect(self.on_link_activated)
        self.setWordWrap(True)

    async def show_song(self, song):
        album = song.album
        artists_str_list = []
        for artist in song.artists:
            artists_str_list.append(
                f'<a href="{reverse(artist)}">{artist.name_display}</a>'
            )
        artists_str = ' / '.join(artists_str_list)
        self.setText(f'''\
专辑名：<a href="{reverse(album)}">{album.name_display}</a>
<br/>
歌手：{artists_str}
''')

    def on_link_activated(self, url):
        model = resolve(url)
        self._app.browser.goto(model=model)


class LeftCon(QWidget, BgTransparentMixin):
    def sizeHint(self):
        size_hint = super().sizeHint()
        return QSize(500, size_hint.height())


class RightCon(QWidget):
    def sizeHint(self):
        size_hint = super().sizeHint()
        return QSize(260, size_hint.height())


class SongExploreView(QWidget):
    def __init__(self, app):
        super().__init__(parent=None)
        self._app = app

        self._left_con = LeftCon(self)  #: left container
        self._left_con_scrollarea = ScrollArea(self._app)
        self._left_con_scrollarea.setWidget(self._left_con)
        self._right_con = RightCon(self)
        self._left_layout = QVBoxLayout(self._left_con)
        self._left_h1_layout = QHBoxLayout()
        self._left_h2_layout = QHBoxLayout()
        self._right_layout = QVBoxLayout(self._right_con)

        self.header_label = HeaderLabel()
        self.lyric_header = HeaderLabel('<h3>歌词：</h3>')
        self.comments_header = HeaderLabel('<h2>热门评论</h2>')
        self.similar_songs_header = HeaderLabel('<h2>相似歌曲</h2>')
        self.lyric_label = LyricLabel()
        self.play_btn = TextButton('播放')
        self.copy_web_url_btn = TextButton('复制网页地址')
        self.cover_label = CoverLabelV2(app=app)
        self.album_info_label = SongAlbumLabel(app)
        self.comments_view = CommentListView(reserved=0)
        self.similar_songs_view = SongListView(reserved=0)

        self._lyric_scrollarea = ScrollArea(self._app)
        self._lyric_scrollarea.setWidget(self.lyric_label)
        self.lyric_label.setSizePolicy(QSizePolicy.Preferred,
                                       QSizePolicy.Expanding)

        self.header_label.setTextFormat(Qt.RichText)
        self.cover_label.setSizePolicy(QSizePolicy.Preferred,
                                       QSizePolicy.Preferred)

        self._layout = QHBoxLayout(self)
        self._setup_ui()

        self.similar_songs_view.play_song_needed.connect(
            app.playlist.play_model)

    def _setup_ui(self):
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(10, 0, 10, 0)

        self._layout.addWidget(self._left_con_scrollarea, 2)
        # Assume the max screen height is about 1080.
        self._right_con.setMaximumWidth(360)
        self._layout.addWidget(self._right_con, 1)

        self._left_layout.addWidget(self.header_label)
        self._left_layout.addLayout(self._left_h1_layout)
        self._left_layout.addLayout(self._left_h2_layout)
        self._left_layout.addWidget(self.similar_songs_header)
        self._left_layout.addWidget(self.similar_songs_view)
        self._left_layout.addWidget(self.comments_header)
        self._left_layout.addWidget(self.comments_view)
        self._left_layout.addSpacing(10)

        self._left_h2_layout.addWidget(self.play_btn)
        self._left_h2_layout.addSpacing(10)
        self._left_h2_layout.addWidget(self.copy_web_url_btn)
        self._left_h2_layout.addStretch(0)
        # self._left_layout.addStretch(0)

        self._right_layout.addWidget(self.cover_label)
        self._right_layout.addWidget(self.album_info_label)
        self._right_layout.addWidget(self.lyric_header)
        self._right_layout.addWidget(self._lyric_scrollarea)
        # self._right_layout.addStretch(0)



from feeluown.server import Request
from .lexer import furi_re, integer_re, float_re


def options_to_str(options):
    """
    TODO: support complex value, such as list
    """
    return ",".join(f"{k}={v}"
                    for k, v in options.items())


def unparse(request: Request):
    """Generate source code for the request object"""

    def escape(value):
        # if value is not furi/float/integer, than we surround the value
        # with double quotes
        regex_list = (furi_re, float_re, integer_re)
        for regex in regex_list:
            if regex.match(value):
                break
        else:
            value = f'"{value}"'
        return value

    # TODO: allow heredoc and args appear at the same time
    args_str = '' if request.has_heredoc else \
        ' '.join((escape(arg) for arg in request.cmd_args))
    options_str = options_to_str(request.cmd_options)
    if options_str:
        options_str = f'[{options_str}]'
    raw = f'{request.cmd} {args_str} {options_str} #: {options_to_str(request.options)}'
    if request.has_heredoc:
        word = request.heredoc_word
        raw += f' <<{word}\n{request.cmd_args[0]}\n{word}\n\n'
    return raw



from .cmd import Cmd
from .base import AbstractHandler


class SetHandler(AbstractHandler):
    cmds = ('set', )

    def handle(self, cmd: Cmd):
        assert self.session is not None
        options = cmd.options

        for key, value in options.items():
            # Treat None as the default value.
            # TODO: use a more meaningful default value.
            if value is not None:
                setattr(self.session.options, key, value)



import asyncio
import logging
import warnings
from contextlib import suppress

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QImage, QPixmap
from PyQt5.QtWidgets import QFrame, QVBoxLayout, QLabel, QApplication
from requests.exceptions import RequestException

from feeluown.utils import aio
from feeluown.utils.reader import wrap
from feeluown.media import Media, MediaType
from feeluown.excs import ProviderIOError
from feeluown.library import ModelState, NotSupported, ModelFlags, ModelType

from feeluown.gui.helpers import BgTransparentMixin, \
    disconnect_slots_if_has, fetch_cover_wrapper
from feeluown.gui.widgets.menu import SongMenuInitializer
from feeluown.gui.widgets.imglist import ImgListView
from feeluown.gui.widgets.album import AlbumListModel, AlbumListView, \
    AlbumFilterProxyModel
from feeluown.gui.widgets.artist import ArtistListModel, ArtistListView, \
    ArtistFilterProxyModel
from feeluown.gui.widgets.video_list import VideoListModel, VideoListView, \
    VideoFilterProxyModel
from feeluown.gui.widgets.playlist import PlaylistListModel, PlaylistListView, \
    PlaylistFilterProxyModel
from feeluown.gui.widgets.songs import ColumnsMode, SongsTableModel, SongsTableView, \
    SongFilterProxyModel
from feeluown.gui.widgets.comment_list import CommentListView, CommentListModel
from feeluown.gui.widgets.meta import TableMetaWidget
from feeluown.gui.widgets.table_toolbar import SongsTableToolbar
from feeluown.gui.widgets.tabbar import TableTabBarV2

logger = logging.getLogger(__name__)


class Renderer:
    async def setUp(self, container):
        # pylint: disable=attribute-defined-outside-init
        self.container = container
        self.meta_widget = container.meta_widget
        self.desc_widget = container.desc_widget
        self.toolbar = container.toolbar
        self.tabbar = container.tabbar
        self.songs_table = container.songs_table
        self.albums_table = container.albums_table
        self.artists_table = container.artists_table
        self.videos_table = container.videos_table
        self.playlists_table = container.playlists_table
        self.comments_table = container.comments_table
        # pylint: disable=protected-access
        self._app = container._app

    async def render(self):
        """render contents in table container

        please follow the following rendering order:

        1. show meta widget and basic metadata, bind signal if needed
        2. fetch data and show content in table, bind signal if needed
        3. fetch description and show, bind signal if needed
        """

    async def tearDown(self):
        pass

    #
    # utils function for renderer
    #
    def set_extra(self, extra):
        self.container.current_extra = extra

    async def show_cover(self, cover, cover_uid, as_background=False):
        cover = Media(cover, MediaType.image)
        url = cover.url
        app = self._app
        content = await app.img_mgr.get(url, cover_uid)
        img = QImage()
        img.loadFromData(content)
        pixmap = QPixmap(img)
        if not pixmap.isNull():
            if as_background:
                self.meta_widget.set_cover_pixmap(None)
                self._app.ui.right_panel.show_background_image(pixmap)
            else:
                self._app.ui.right_panel.show_background_image(None)
                self.meta_widget.set_cover_pixmap(pixmap)
            self._app.ui.table_container.updateGeometry()

    def show_albums(self, reader):
        self._show_model_with_cover(reader,
                                    self.albums_table,
                                    AlbumListModel,
                                    AlbumFilterProxyModel)

    def show_artists(self, reader):
        self._show_model_with_cover(reader,
                                    self.artists_table,
                                    ArtistListModel,
                                    ArtistFilterProxyModel)

    def show_videos(self, reader):
        self._show_model_with_cover(reader,
                                    self.videos_table,
                                    VideoListModel,
                                    VideoFilterProxyModel)

    def show_playlists(self, reader):
        self._show_model_with_cover(reader,
                                    self.playlists_table,
                                    PlaylistListModel,
                                    PlaylistFilterProxyModel)

    def _show_model_with_cover(self, reader, table, model_cls, filter_model_cls):
        self.container.current_table = table
        filter_model = filter_model_cls()
        source_name_map = {p.identifier: p.name for p in self._app.library.list()}
        model = model_cls(reader,
                          fetch_cover_wrapper(self._app),
                          source_name_map=source_name_map)
        filter_model.setSourceModel(model)
        table.setModel(filter_model)
        table.scrollToTop()
        disconnect_slots_if_has(self._app.ui.magicbox.filter_text_changed)
        self._app.ui.magicbox.filter_text_changed.connect(filter_model.filter_by_text)

    def show_songs(self, reader, show_count=False, columns_mode=ColumnsMode.normal):
        """
        .. versionadded: v3.8.5
           The *hide_columns* parameter.
        """
        reader = wrap(reader)
        if show_count:
            count = reader.count
            self.meta_widget.songs_count = -1 if count is None else count
        source_name_map = {p.identifier: p.name for p in self._app.library.list()}
        model = SongsTableModel(
            source_name_map=source_name_map,
            reader=reader,
            parent=self.songs_table,
        )
        self.show_songs_by_model(model, columns_mode=columns_mode)

    def show_songs_by_model(self, model, columns_mode=ColumnsMode.normal):
        self.container.current_table = self.songs_table
        self.toolbar.show()
        filter_model = SongFilterProxyModel(self.songs_table)
        filter_model.setSourceModel(model)
        self.songs_table.setModel(filter_model)
        self.songs_table.scrollToTop()
        self.songs_table.set_columns_mode(columns_mode)
        disconnect_slots_if_has(self._app.ui.magicbox.filter_text_changed)
        self._app.ui.magicbox.filter_text_changed.connect(filter_model.filter_by_text)

    def show_desc(self, desc):
        self.container.current_table = None
        self.desc_widget.setText(desc)
        self.desc_widget.show()

    def show_comments(self, comments):
        self.container.current_table = self.comments_table
        reader = wrap(comments)
        model = CommentListModel(reader)
        self.comments_table.setModel(model)

    def _get_source_alias(self, source):
        provider = self._app.library.get(source)
        return provider.name if provider is not None else ''


class SongsCollectionRenderer(Renderer):
    def __init__(self, collection):
        self.collection = collection

    async def render(self):
        collection = self.collection
        self.meta_widget.show()
        self.meta_widget.title = collection.name
        self.meta_widget.updated_at = collection.updated_at
        self.meta_widget.created_at = collection.created_at
        self.songs_table.remove_song_func = self.remove_song
        self._show_songs()

        # only show tabbar if description is not empty
        if self.collection.description:
            self.tabbar.show()
            # FIXME: maybe add a playlist/collection mode for tabbar
            self.tabbar.album_mode()
            self.tabbar.show_songs_needed.connect(self._show_songs)
            self.tabbar.show_desc_needed.connect(
                lambda: self.show_desc(self.collection.description))

    def _show_songs(self):
        """filter model with other type"""
        self.show_songs(wrap([model for model in self.collection.models
                              if model.meta.model_type == ModelType.song]))

    def remove_song(self, song):
        self.collection.remove(song)
        # Re-render songs table so that user can see that the song is removed.
        self._show_songs()


class AlbumsCollectionRenderer(Renderer):
    def __init__(self, reader):
        self.reader = reader

    async def render(self):
        # always bind signals first
        self.toolbar.filter_albums_needed.connect(
            lambda types: self.albums_table.model().filter_by_types(types))

        self.show_albums(self.reader)


class ArtistsCollectionRenderer(Renderer):
    def __init__(self, reader):
        self.reader = reader

    async def render(self):
        self.show_artists(self.reader)


class VideosRenderer(Renderer):
    def __init__(self, reader):
        self.reader = reader

    async def render(self):
        self.show_videos(self.reader)


class DescLabel(QLabel):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setContentsMargins(30, 15, 30, 10)
        self.setWordWrap(True)
        self.setTextInteractionFlags(Qt.TextSelectableByMouse)


class TableContainer(QFrame, BgTransparentMixin):
    def __init__(self, app, parent=None):
        super().__init__(parent)

        self._app = app
        self._renderer = None
        self._table = None  # current visible table
        self._tables = []

        self._extra = None
        self.toolbar = SongsTableToolbar()
        self.tabbar = TableTabBarV2()
        self.meta_widget = TableMetaWidget(parent=self)
        self.songs_table = SongsTableView(parent=self)
        self.albums_table = AlbumListView(parent=self)
        self.artists_table = ArtistListView(parent=self)
        self.videos_table = VideoListView(parent=self)
        self.playlists_table = PlaylistListView(parent=self)
        self.comments_table = CommentListView(parent=self)
        self.desc_widget = DescLabel(parent=self)

        self._tables.append(self.songs_table)
        self._tables.append(self.albums_table)
        self._tables.append(self.artists_table)
        self._tables.append(self.playlists_table)
        self._tables.append(self.videos_table)
        self._tables.append(self.comments_table)

        self.songs_table.play_song_needed.connect(
            lambda song: aio.run_afn(self.play_song, song))
        self.videos_table.play_video_needed.connect(
            self._app.playlist.play_model)

        def goto_model(model):
            self._app.browser.goto(model=model)

        for signal in [self.songs_table.show_artist_needed,
                       self.songs_table.show_album_needed,
                       self.albums_table.show_album_needed,
                       self.artists_table.show_artist_needed,
                       self.playlists_table.show_playlist_needed,
                       ]:
            signal.connect(goto_model)

        self.toolbar.play_all_needed.connect(
            lambda: aio.run_afn(self.play_all))
        self.songs_table.add_to_playlist_needed.connect(self._add_songs_to_playlist)
        self.songs_table.about_to_show_menu.connect(self._songs_table_about_to_show_menu)
        self.songs_table.activated.connect(
            lambda index: aio.create_task(self._on_songs_table_activated(index)))

        self._setup_ui()

    def _setup_ui(self):
        self.current_table = None
        self.tabbar.hide()
        self.meta_widget.add_tabbar(self.tabbar)
        self.desc_widget.hide()

        self._layout = QVBoxLayout(self)
        self._v_layout = QVBoxLayout()

        self._v_layout.addWidget(self.meta_widget)
        self._v_layout.addWidget(self.toolbar)
        self._v_layout.addSpacing(10)
        self._v_layout.addWidget(self.desc_widget)

        # Since QTableView has a margin on the left and right(see SongsTableDelegate),
        # so set the v_layout left margin to same value(0) to align widgets.
        self._v_layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addLayout(self._v_layout)
        for table in self._tables:
            self._layout.addWidget(table)
        self._layout.addStretch(0)
        self._layout.setContentsMargins(20, 0, 20, 0)
        self._layout.setSpacing(0)

    @property
    def current_extra(self):
        return self._extra

    @current_extra.setter
    def current_extra(self, extra):
        """(alpha)"""
        if self._extra is not None:
            self._layout.removeWidget(self._extra)
            self._extra.deleteLater()
            del self._extra
        self._extra = extra
        if self._extra is not None:
            self._layout.insertWidget(1, self._extra)

    @property
    def current_table(self):
        """current visible table, if no table is visible, return None"""
        return self._table

    @current_table.setter
    def current_table(self, table):
        """set table as current visible table

        show table and hide other tables, if table is None,
        hide all tables.
        """
        for t in self._tables:
            if t != table:
                t.hide()
        if table is None:
            self.toolbar.hide()
        else:
            self.desc_widget.hide()
            table.show()
            if table is self.artists_table:
                self.toolbar.artists_mode()
            if table is self.albums_table:
                self.toolbar.albums_mode()
            if table is self.songs_table:
                self.toolbar.songs_mode()
        if isinstance(self._table, ImgListView):
            self._table.setModel(None)
        self._table = table

    async def set_renderer(self, renderer):
        """set ui renderer

        TODO: add lock for set_renderer
        """

        if renderer is None:
            return

        # firstly, tear down everything
        # tear down last renderer
        if self._renderer is not None:
            await self._renderer.tearDown()
        self.meta_widget.hide()
        self.meta_widget.clear()
        self.tabbar.hide()
        self.tabbar.restore_default()
        self.current_table = None
        self.current_extra = None
        # clean right_panel background image
        self._app.ui.right_panel.show_background_image(None)
        # disconnect songs_table signal
        signals = (
            self.tabbar.show_contributed_albums_needed,
            self.tabbar.show_albums_needed,
            self.tabbar.show_songs_needed,
            self.tabbar.show_artists_needed,
            self.tabbar.show_playlists_needed,
            self.tabbar.show_desc_needed,
        )
        for signal in signals:
            disconnect_slots_if_has(signal)

        # unbind some callback function
        self.songs_table.remove_song_func = None

        # secondly, prepare environment
        self.show()

        # thirdly, setup new renderer
        await renderer.setUp(self)
        self._renderer = renderer
        await self._renderer.render()

    async def play_song(self, song):
        self._app.playlist.play_model(song)

    async def play_all(self):
        task_name = 'play-all'
        task_spec = self._app.task_mgr.get_or_create(task_name)
        model = self.songs_table.model()
        # FIXME: think about a more elegant way to get reader.
        reader = model.sourceModel().reader
        if reader is not None and reader.count is not None:
            self.toolbar.enter_state_playall_start()
            with suppress(ProviderIOError, asyncio.CancelledError):
                songs = await task_spec.bind_blocking_io(reader.readall)
                self._app.playlist.set_models(songs)
                task = self._app.playlist.next()
                if task is not None:
                    await task
                    self._app.player.resume()
            self.toolbar.enter_state_playall_end()
            return
        assert False, 'The play_all_btn should be hide in this page'

    def show_collection(self, coll):
        renderer = SongsCollectionRenderer(coll)
        aio.create_task(self.set_renderer(renderer))

    def show_songs(self, songs=None, songs_g=None):
        """(DEPRECATED) provided only for backward compatibility"""
        warnings.warn('use readerer.show_songs please')
        renderer = Renderer()
        task = aio.create_task(self.set_renderer(renderer))
        if songs is not None:
            reader = wrap(songs)
        else:
            reader = songs_g
        task.add_done_callback(
            lambda _: renderer.show_songs(reader=reader))

    def show_albums_coll(self, albums_g):
        warnings.warn('use readerer.show_albums please')
        aio.create_task(self.set_renderer(AlbumsCollectionRenderer(albums_g)))

    def show_artists_coll(self, artists_g):
        warnings.warn('use readerer.show_artists please')
        aio.create_task(self.set_renderer(ArtistsCollectionRenderer(artists_g)))

    def search(self, text):
        if self.isVisible() and self.songs_table is not None:
            self.songs_table.filter_row(text)

    def _add_songs_to_playlist(self, songs):
        for song in songs:
            self._app.playlist.add(song)

    def _songs_table_about_to_show_menu(self, ctx):
        models = ctx['models']
        if not models or models[0].meta.model_type != ModelType.song:
            return

        menu = ctx['menu']
        song = models[0]
        menu.addSeparator()
        SongMenuInitializer(self._app, song).apply(menu)

    async def _on_songs_table_activated(self, index):
        """
        QTableView should have no IO operations.
        """
        from feeluown.gui.widgets.songs import Column

        song = index.data(Qt.UserRole)
        if index.column() == Column.song:
            # .. versionadded:: 3.7.11
            #    Reset playlist with songs in table if Alt key is pressed.
            if QApplication.keyboardModifiers() & Qt.AltModifier:
                model = self.songs_table.model()
                if isinstance(model, SongFilterProxyModel):
                    model = model.sourceModel()
                    if isinstance(model, SongsTableModel):
                        reader = model.reader
                        songs = await aio.run_fn(reader.readall)
                        self._app.playlist.clear()
                        self._app.playlist.set_models(songs)
            self._app.playlist.play_model(song)
        else:
            try:
                song = await aio.run_in_executor(
                    None, self._app.library.song_upgrade, song)
            except NotSupported:
                assert ModelFlags.v2 & song.meta.flags
                self._app.show_msg('资源提供放不支持该功能')
                logger.info(f'provider:{song.source} does not support song_get')
                song.state = ModelState.cant_upgrade
            except (ProviderIOError, RequestException) as e:
                # FIXME: we should only catch ProviderIOError here,
                # but currently, some plugins such fuo-qqmusic may raise
                # requests.RequestException
                logger.exception('upgrade song failed')
                self._app.show_msg(f'请求失败: {str(e)}')
            else:
                if index.column() == Column.artist:
                    artists = song.artists
                    if artists:
                        if len(artists) > 1:
                            self.songs_table.show_artists_by_index(index)
                        else:
                            self.songs_table.show_artist_needed.emit(artists[0])
                elif index.column() == Column.album:
                    self.songs_table.show_album_needed.emit(song.album)
        model = self.songs_table.model()
        topleft = model.index(index.row(), 0)
        bottomright = model.index(index.row(), 4)
        model.dataChanged.emit(topleft, bottomright, [])



from PyQt5.QtCore import QMetaObject, pyqtSlot
from PyQt5.QtOpenGL import QGLContext

# HELP: 需要 import GL 模块，否则在 Linux(Ubuntu 18.04) 下会出现 seg fault
from OpenGL import GL  # noqa

from mpv_old import _mpv_get_sub_api, _mpv_opengl_cb_set_update_callback, \
        _mpv_opengl_cb_init_gl, OpenGlCbGetProcAddrFn, _mpv_opengl_cb_draw, \
        _mpv_opengl_cb_report_flip, MpvSubApi, OpenGlCbUpdateFn, _mpv_opengl_cb_uninit_gl

from feeluown.gui.widgets.video import VideoOpenGLWidget


def get_proc_addr(_, name):
    glctx = QGLContext.currentContext()
    if glctx is None:
        return 0
    addr = int(glctx.getProcAddress(name.decode('utf-8')))
    return addr


class MpvOpenGLWidget(VideoOpenGLWidget):
    """Mpv 视频输出窗口

    销毁时，应该调用 shutdown 方法来释放资源。

    该 Widget 是模仿一些 C++ 程序编写而成（详见项目 research 目录下的例子），
    我们对背后的原理不太理解，目前测试是可以正常工作的。

    目前主要的疑问有：

    - [ ] shutdown 方法期望是用来释放资源的，目前的写法不知是否合理？
          应该在什么时候调用 shutdown 方法？
    """

    def __init__(self, app, parent=None):
        super().__init__(app=app, parent=parent)
        self._app = app
        self.mpv_gl = _mpv_get_sub_api(app.player._mpv.handle,
                                       MpvSubApi.MPV_SUB_API_OPENGL_CB)
        self.on_update_c = OpenGlCbUpdateFn(self.on_update)
        self.on_update_fake_c = OpenGlCbUpdateFn(self.on_update_fake)
        self.get_proc_addr_c = OpenGlCbGetProcAddrFn(get_proc_addr)
        self.frameSwapped.connect(self.swapped)

        self._mpv_gl_inited = False

    def shutdown(self):
        if self._mpv_gl_inited:
            self.makeCurrent()
            if self.mpv_gl:
                _mpv_opengl_cb_set_update_callback(
                    self.mpv_gl, self.on_update_fake_c, None)
            _mpv_opengl_cb_uninit_gl(self.mpv_gl)
            self.doneCurrent()

    def initializeGL(self):
        _mpv_opengl_cb_init_gl(self.mpv_gl, None, self.get_proc_addr_c, None)
        _mpv_opengl_cb_set_update_callback(self.mpv_gl, self.on_update_c, None)
        self._mpv_gl_inited = True
        self.context().aboutToBeDestroyed.connect(self.shutdown)

    def paintGL(self):
        # compatible with HiDPI display
        ratio = self.context().screen().devicePixelRatio()
        w = int(self.width() * ratio)
        h = int(self.height() * ratio)
        _mpv_opengl_cb_draw(self.mpv_gl, self.defaultFramebufferObject(), w, -h)

    @pyqtSlot()
    def maybe_update(self):
        if self.window().isMinimized():
            self.makeCurrent()
            self.paintGL()
            self.context().swapBuffers(self.context().surface())
            self.swapped()
            self.doneCurrent()
        else:
            self.update()

    def on_update(self, ctx=None):
        # HELP: maybeUpdate 中的部分逻辑需要在主线程中执行，
        # 而 QMetaObject.invokeMethod 似乎正好可以达到这个目标。
        # 我们将 maybe_update 标记为 pyqtSlot，这样才能正确的 invoke。
        QMetaObject.invokeMethod(self, 'maybe_update')

    def on_update_fake(self, ctx=None):
        pass

    def swapped(self):
        _mpv_opengl_cb_report_flip(self.mpv_gl, 0)

    def closeEvent(self, _):
        self.shutdown()


# TODO: 实现 MpvEmbeddedWidget
class MpvEmbeddedWidget:
    pass



#!/usr/bin/env python3

from setuptools import setup, find_packages
from os import path

import feeluown

current_directory = path.abspath(path.dirname(__file__))
with open(path.join(current_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='feeluown',
    version=feeluown.__version__,
    description='*nix music player',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="GPL-3.0",
    author='feeluown',
    author_email='yinshaowen241@gmail.com',
    packages=find_packages(exclude=('tests*',)),
    py_modules=['mpv', 'mpv_old'],
    package_data={
        '': ['nowplaying/linux/*.xml',
             'gui/assets/icons/*.png',
             'gui/assets/icons/*.ico',
             'gui/assets/icons/*.icns',
             'gui/assets/themes/*.qss',
             'gui/assets/themes/*.colors',
             ]
    },
    python_requires=">=3.8",
    url='https://github.com/feeluown/FeelUOwn',
    keywords=['media', 'player', 'application', 'PyQt5', 'Python 3'],
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Environment :: X11 Applications :: Qt',
        "Topic :: Multimedia :: Sound/Audio",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    # FIXME: depends on PyQt5
    # - PyQt5.QtWidgets/QtCore/QtGui
    # - PyQt5.QtSvg
    # - PyQt5.QtOpenGL
    #
    # The feeluown.compat module will choose the right package
    # for different python version
    install_requires=[
        'janus',
        'requests',
        'qasync',
        'tomlkit',
        'packaging',
        'pydantic>=1.8.1',
    ],
    extras_require={
        'battery': ['fuo-local>=0.2.1',
                    'fuo-netease>=0.4.2',
                    'fuo-qqmusic>=0.2',
                    'fuo-kuwo>=0.1.1',
                    ],
        'macOS': ['pyobjc-framework-Cocoa', 'pyobjc-framework-Quartz'],
        'win32': ['pyshortcuts'],
        'webengine': ['PyQtWebEngine'],
        'dev': [
            # lint
            'flake8',
            'pylint',
            'mypy',
            # pyqt5-stubs seems more accurate than the stubs packaged in PyQt5 package.
            # For example, the stubs in PyQt5 has wrong type hint for Qt.ItemIsSelectable.
            # Some pyqt5 package installed by system package manager may not contain stubs.
            'pyqt5-stubs',

            # unittest
            'pytest>=5.4.0',
            'pytest-runner',
            'pytest-cov',
            'pytest-asyncio',
            'pytest-qt',
            'pytest-mock',
            'pytest-benchmark>=3.4.1',

            # docs
            'sphinx',
            'sphinx_rtd_theme',
        ],
    },
    entry_points={
        'console_scripts': [
            "feeluown=feeluown.__main__:main",
            "fuo=feeluown.__main__:main",
            "feeluown-genicon=feeluown.cli.install:generate_icon",
            # "feeluown-update=feeluown.install:update"
        ]
    },
)



from .uimodels.collection import CollectionUiManager
from .uimodels.provider import ProviderUiManager
from .uimodels.my_music import MyMusicUiManager
from .uimodels.playlist import PlaylistUiManager


__all__ = (
    'CollectionUiManager',
    'ProviderUiManager',
    'MyMusicUiManager',
    'PlaylistUiManager',
)



from PyQt5.QtCore import Qt, QObject, QCoreApplication, QEvent
from PyQt5.QtGui import QKeySequence as KS, QMouseEvent
from PyQt5.QtWidgets import QShortcut


class HotkeyManager(QObject):
    """集中管理各个组件的快捷键"""

    def __init__(self, app):
        super().__init__(app)
        self._app = app

    def initialize(self):
        app = self._app
        ui = self._app.ui

        # magicbox
        QShortcut(KS('Ctrl+F'), app).activated.connect(
            ui.toolbar.show_and_focus_magicbox)

        # player
        QShortcut(KS(Qt.Key_Space), app).activated.connect(app.player.toggle)

        def p_shortcut_connect(k, cb):  # an alias to simplify code
            sc = QShortcut(KS(k), ui.pc_panel)
            sc.setContext(Qt.WidgetWithChildrenShortcut)
            sc.activated.connect(cb)

        p_shortcut_connect(Qt.Key_Right, self._player_forward_a_little)
        p_shortcut_connect(Qt.Key_Left, self._player_backward_a_little)
        p_shortcut_connect(Qt.Key_Up, self._player_volume_up_a_little)
        p_shortcut_connect(Qt.Key_Down, self._player_volume_down_a_little)

        # browser
        QShortcut(KS.Back, app).activated.connect(app.browser.back)
        QShortcut(KS.Forward, app).activated.connect(app.browser.forward)

        # install event filter on app
        q_app = QCoreApplication.instance()
        q_app.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseButtonPress:
            # The event can be QContextMenuEvent.
            if not isinstance(event, QMouseEvent):
                return False
            button = event.button()
            if button == Qt.BackButton:
                self._app.browser.back()
                return True
            elif button == Qt.ForwardButton:
                self._app.browser.forward()
                return True
        return False

    def _player_forward_a_little(self):
        old_position = self._app.player.position
        duration = self._app.player.duration
        if None not in (old_position, duration):
            self._app.player.position = min(duration - 1, old_position + 5)

    def _player_backward_a_little(self):
        old_position = self._app.player.position
        if old_position is not None:
            self._app.player.position = max(0, old_position - 5)

    def _player_volume_up_a_little(self):
        self._app.player.volume = min(100,
                                      self._app.player.volume + 10)

    def _player_volume_down_a_little(self):
        self._app.player.volume = max(0,
                                      self._app.player.volume - 10)



from PyQt5.QtCore import Qt, QModelIndex

from feeluown.utils import aio
from feeluown.utils.reader import create_reader
from feeluown.gui.widgets import TextButton
from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.gui.page_containers.table import Renderer
from feeluown.gui.widgets.songs import BaseSongsTableModel, Column, ColumnsMode


async def render(req, **kwargs):
    """/player_playlist handler
    """
    app = req.ctx['app']
    tab_name = req.query.get('tab_name', '')

    right_panel = app.ui.right_panel
    right_panel.set_body(right_panel.table_container)
    aio.create_task(
        app.ui.table_container.set_renderer(PlayerPlaylistRenderer(tab_name)))


class PlayerPlaylistRenderer(Renderer, TabBarRendererMixin):

    def __init__(self, tab_name):
        self.tabs = [
            # (title, tab_name, )
            ('播放列表', 'playlist', ),
            ('最近播放', 'recently_played', )
        ]
        self.tab_index = self._get_tabindex_by_tabname(tab_name) or 0

    async def render(self):
        self.meta_widget.title = self.tabs[self.tab_index][0]
        self.meta_widget.show()
        self.render_tab_bar()

        if self.tab_index == 0:
            await self.render_playlist()
        else:
            await self.render_recently_played()

    async def render_recently_played(self):
        songs = self._app.recently_played.list_songs()
        reader = create_reader(songs)
        self.show_songs(reader)
        self.toolbar.hide()

    async def render_playlist(self):
        self.songs_table.remove_song_func = self._app.playlist.remove
        source_name_map = {p.identifier: p.name for p in self._app.library.list()}
        model = PlaylistTableModel(self._app.playlist, source_name_map)
        self.show_songs_by_model(model, columns_mode=ColumnsMode.playlist)

        # TODO(cosven): the toolbar is useless, and we should remove it lator.
        self.toolbar.manual_mode()
        btn = TextButton('清空', self.toolbar)
        btn.clicked.connect(lambda *args: aio.create_task(self.clear_playlist()))
        self.toolbar.add_tmp_button(btn)

        # Scroll to current song.
        current_song = self._app.playlist.current_song
        if current_song is not None:
            row = self._app.playlist.list().index(current_song)
            model_index = self.songs_table.model().index(row, Column.song)
            self.songs_table.scrollTo(model_index)
            self.songs_table.selectRow(row)

    async def clear_playlist(self):
        self._app.playlist.clear()
        await self.render()  # re-render

    def render_by_tab_index(self, tab_index):
        tab_name = self.tabs[tab_index][1]
        self._app.browser.goto(page='/player_playlist',
                               query={'tab_name': tab_name})

    def _get_tabindex_by_tabname(self, tab_name):
        for i, tab in enumerate(self.tabs):
            if tab[1] == tab_name:
                return i
        return 0


class PlaylistTableModel(BaseSongsTableModel):
    def __init__(self, playlist, source_name_map=None, parent=None):
        """

        :param songs: 歌曲列表
        :param songs_g: 歌曲列表生成器（当歌曲列表生成器不为 None 时，忽略 songs 参数）
        """
        super().__init__(source_name_map, parent)
        self._playlist = playlist
        self._items = playlist.list().copy()

        self._playlist.songs_added.connect(self.on_songs_added)
        self._playlist.songs_removed.connect(self.on_songs_removed)

    def flags(self, index):
        flags = super().flags(index)
        if index.column() == Column.song:
            song = index.data(Qt.UserRole)
            if self._playlist.is_bad(song):
                # a bad song is disabled
                flags &= ~Qt.ItemIsEnabled
        return flags

    def on_songs_added(self, index, count):
        self.beginInsertRows(QModelIndex(), index, index+count-1)
        # Insert from tail to front.
        while count > 0:
            self._items.insert(index, self._playlist[index+count-1])
            count -= 1
        self.endInsertRows()

    def on_songs_removed(self, index, count):
        self.beginRemoveRows(QModelIndex(), index, index+count-1)
        while count > 0:
            self._items.pop(index + count - 1)
            count -= 1
        self.endRemoveRows()
